;;; www-index.el
;;; Michael Ernst <mernst@research.microsoft.com>
;;; Last modified 3/95

;; This file automates indexing proper names in HTML documents.  The
;; indexing scheme is described below.
;; The top-level functions are:
;;   process-indexees-for-{file,directory}
;;       Add indexing commands to a file or every HTML file in a directory
;;       See indexing-keystroke-help for what user input is expected.
;;   {read,write}-index-obarray
;;       Read or write the database of abbreviations.
;;       It is read when this file is loaded, but you must write it yourself.
;;       Variable index-obarray-file controls the file in which it is saved.

;; To do:
;;  * make hyphens work:  Port-au-Prince is a single proper noun.
;; Postprocessing:
;;  * if many entries, combine into a single concept entry.
;;  * generate "see" or "seealso" when index entry differs from enclosed
;;      text, isn't abbreviation or prefix, isn't a person.  (Maybe do this
;;      in Perl instead.)

(provide 'www-index)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Indexing scheme
;;; 

;; Types of index entry:
;; 
;; <a href="(../)*index-[a-z].html#...(|flags)?>...</a>
;;       also <!-- a href="..." -->...<!-- /a -->
;;  flags (follow vertical bar in index entry):
;;   q	Index item should be surrounded by double-quote marks
;;   i	Index item should be italicized
;;   c	Index item belongs in concept index as well.
;;	This should only appear in a special index bar at article bottom (?).
;;   o	One-way reference: point at index, but don't be pointed back to by it.
;;	This is good for side comments, off-the-subject comparisons, etc.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Add index entries
;;;

;; We need to do two things simultaneously:
;;  1. Select a word or group of words of interest.
;;  2. Classify the word, and process it.

;; First find a likely-looking set of words.  If they have already been
;; classified, as a y-or-n-p question to see whether this set is the same
;; usage.  (If there are multiple classifications, first ask about the most
;; recently seen.)

;; Read this help:
(defvar indexing-keystroke-help "\
***************************************************************************
To index a phrase:
  h     human being (H: alias for human)
  o     other proper noun, eg organization (N: alias for organization)
  p     whole publication, eg book or magazine (P: alias for publication)
  q     part of a publication typeset quoted, eg article (Q: alias for article)

To ignore a phrase:
  n,DEL don't index; reconsider starting at second work
  N     don't index, and go to end of group before looking again
  x     ignore this word for the rest of the file
  s     skip remainder of line
  !     skip rest of file

To select the phrase:
  C-a, C-e    subsequent actions affect front (end) of word group
  C-f, C-b    adjust word group boundary by one character
  M-f, M-b    adjust word group boundary by one word  (M-DEL acts like M-b)
***************************************************************************
")

(defvar index-obarray nil
  "obarray of symbols for classified indexable items.
The value of each symbol is a list of classifications.")

;;; Each classification is one of:
;; ?h  human being
;; ?o  organization
;; ?p  publication
;; ?q  article or other quoted part of a publication
;; (?H . name)  alias for human
;; (?O . name)  alias for organization
;; (?P . name)  alias for publication
;; (?Q . name)  alias for article

;; When one of these words is the only capitalized word and it appears at
;; the beginning of a sentence, skip over it.
(defvar sentence-start-false-capitalizations
  '("A" "About" "According" "Accordingly" "Across" "Activists" "Actually"
    "Adding" "After" "All" "Almost" "Also" "Although"
    "Among" "An" "And" "Annual" "Another" "Anti" "Any" "Anyway"
    "Apparently" "Arguably" "As" "Aside" "At"
    "Back" "Because" "Below" "Beneath" "Besides" "Best"
    "Big" "Blaming" "Both" "But" "Buy" "Buying" "By"
    "Can" "Clearly" "Commentators" "Conservative" "Consider" "Contrary" "Corporate" "Coverage" "Critics" "Current" "Currently"
    "Despite" "Devoid" "Did" "Didn't" "Do" "Does" "Doesn't" "Don't" "Due" "During"
    "Earlier" "Early" "Editors" "Eight" "Eighty" "Elsewhere"
    "Even" "Every" "Exactly" "Except"
    "Faced" "Far" "Farther" "Few" "Fifty" "Five" "Finally" "First"
    "Following" "For" "Former" "Forty" "Four" "From" "Furhter" "Furthermore"
    "Get" "Given" "Good"
    "Had" "Hailed" "Has" "Have" "Haven't"
    "He" "He'll" "Headlines" "Hence" "Her" "Here"
    "His" "How" "However" "Huge"
    "I" "I'd" "I'm" "I've" "If" "In" "Indeed" "Indulging" "Initially" "Instead"
    "Ironically" "Is" "Isn't" "It" "It's" "Its"
    "Journalism" "Journalists" "Just"
    "Keep"
    "Last" "Lately" "Later" "Liberal" "Less" "Let" "Like" "Little" "Look"
    "Mainstream" "Maintaining" "Making" "Many" "Maybe"
    "Meanwhile" "Men" "More" "Moreover" "Most" "My"
    "Near" "Nearly" "Needless" "Neither" "Never" "Nevertheless" "Next"
    "Nine" "Ninety"
    "No" "None" "Nonetheless" "Nor" "Not" "Note" "Notice" "Now" "Nowadays" "Nowhere"
    "Occasionally" "Of" "Official" "Officials" "Often"
    "On" "Once" "One" "Only" "Opinion" "Or" "Other" "Otherwise" "Our" "Out" "Over"
    "Part" "Perhaps" "Please" "Polls" "Predictably" "Previous" "Probably" "Public"
    "Reconsideration" "Referring" "Reporters" "Reporting"
    "Scholarly" "Send" "Seven" "Seventy" "Several" "She" "She'll" "Shortly" "Should"
    "Sidebar" "Significantly" "Similarly" "Since" "Six" "Sixty" "Small"
    "So" "Some" "Somehow" "Someone" "Soemtimes" "Special" "Spokespersons" "Students"
    "Such" "Surely" "Syndicated"
    "Taxpayers" "Ten" "That" "The" "Their" "Then" "There" "These"
    "They" "They'll" "They're" "They've" "Thirty" "This" "Those" "Though" "Thus"
    "To" "Three" "Through" "Throughout" "Twenty" "Two" "Typically" "Typical"
    "Under" "Underneath" "Unfortunately" "Unlike" "Up" "Upon" "Usually"
    "Was" "We" "We're" "Welfare" "We'll" "Well"
    "What" "When" "Where" "Whether" "Which" "While" "Who" "Why"
    ;; Not Will; it's a last name for George Will
    "With" "Within" "Without" "Women" "Worse" "Worst" "Writing"
    "Very" "Violence"
    "Yet" "You" "Your"))

(defvar always-ignore-capitalizations
  '("Best of EXTRA" "EXTRA" "FAIR" "I" "I'd" "I'm" "I've"
    "Media Beat" "TV" "U.S."))

(defvar month-regexp "\\(Jan\\|January\\|Feb\\|February\\|Mar\\|March\\|Apr\\|April\\|May\\|June?\\|July?\\|Aug\\|August\\|Sept?\\|September\\|Oct\\|October\\|Nov\\|November\\|Dec\\|December\\)\\.?")

(defun find-potential-indexee ()
  "Search forward from point for a candidate word or phrase for indexing.
Return a cons of the buffer points bounding the word or phrase, or nil
if no candidate is found."
  (let ((case-fold-search nil))
    (if (re-search-forward "\"?\\<[A-Z]" nil t)
	(progn
	  (let* ((start-point (match-beginning 0))
		 end-point)
	    (while (progn
		     (if (not (= (char-after (point)) ?-))
			 (backward-char 1))
		     (forward-word 1)
		     (if (and (= (char-after (point)) ?.)
			      ;; One-letter abbreviation (U.S.) or title
			      (or (looking-back-at "[\\. \n][A-Z]")
				  (looking-back-at "Mr\\|Mrs\\|Ms\\|Prof\\|St")))
			 (forward-char 1))
		     ;; Abbreviations (U.S., N.A.S.A.)
		     (if (looking-at "\\.\\([A-Z]\\.\\)+")
			 (goto-char (match-end 0)))
		     (setq end-point (point))
		     (skip-interword-space-forward)
		     (if (looking-at
			  "\\(a\\|and\\|for\\|in\\|of\\|of the\\|on\\|the\\|to\\)\\>")
			 (progn
			   (forward-word 1)
			   (skip-interword-space-forward)))
		     (if (and (looking-at "([A-Z][A-Z]+)")
			      (not (looking-at "(\\(PBS\\|NBC\\|CNN\\|ABC\\))")))
			 (progn
			   (forward-sexp 1)
			   (setq end-point (point))
			   (skip-interword-space-forward)))
		     (if (looking-at "/[A-Z]")
			 (forward-char 1))
		     (if (looking-at "&[ \n]World[ \n]Report")
			 (progn
			   (goto-char (match-end 0))
			   (setq end-point (point))
			   (skip-interword-space-forward)))
		     (looking-at "[A-Z]"))
	      nil)
	    (goto-char end-point)
	    (if (= (char-after start-point) ?\")
		(if (looking-at "[,.;!?]?\"")
		    (goto-char (match-end 0))
		  (setq start-point (1+ start-point))))
	    (if (looking-back-at "'s")
		(backward-char 2))
	    (if (and (= (preceding-char) ?')
		     (not (member (char-after start-point) '(?' ?`))))
		(backward-char 1))
	    (setq end-point (point))

	    ;; Trim potential matches.
	    (let ((potential-indexee (buffer-substring start-point end-point)))
	      (if (or
		   ;; A capitalized word or phrase we never index
		   (member potential-indexee always-ignore-capitalizations)
		   ;; A capitalized word we ignore at sentence beginning
		   (save-excursion
		     (and
		      ;; one word
		      (progn (backward-word 1)
			     (or (= (point) start-point)
				 (= (point) (1+ start-point))))
		      ;; at beginning of sentence
		      (or
		       (looking-back-at
			(concat
			 "\\(\">\\|<br>\\|<li>\\|\n\\|[.!?][\]'\"\)]*\\|:\\)"
			 "\\([ \t]*\n\\)?[ \t]*[\[\(\"']*"))
		       (looking-back-at ",\\([ \t]*\n\\)?[ \t]*[\[\(\"']*"))
		      (or (member potential-indexee
				  sentence-start-false-capitalizations)
			  (string-match "^According to" potential-indexee))))
		   ;; A month
		   (and
		    (looking-at "[.,]?[ \t\n]*'?[1-9]")
		    (string-match (concat "^\\(\\(In\\|Last\\|On\\|The\\)[ \n]\\)?"
					  month-regexp "$")
				  potential-indexee)))
		  ;; This candidate failed; recursive call to find another
		  (find-potential-indexee)
		(cons start-point end-point))))))))

(defun skip-interword-space-forward ()
  (skip-chars-forward " \t")
  (if (eq (char-after (point)) ?\n)
      (forward-char 1)
    ;; hyphen must be quoted
    (skip-chars-forward "\\-"))
  (skip-chars-forward " \t"))

(defun skip-interword-space-backward ()
  (skip-chars-backward " \t")
  (if (eq (preceding-char) ?\n)
      (backward-char 1))
  (skip-chars-backward " \t"))

(defun clean-up-whitespace (string)
  (while (string-match "[\t\n]" string)
    (aset string (match-beginning 0) ?\ ))
  (while (string-match "  +" string)
    (setq string (concat (substring string 0 (1+ (match-beginning 0)))
			 (substring string (match-end 0)))))
  string)
;; (clean-up-whitespace "foo \n \t bar")

(defvar dynamic-top-directory nil
  "Top directory, relative to current file.  Bound dynamically.")

;; (defvar top-directory-prefix "/u/mernst/WWW/fair/")
(defvar top-directory-prefix "/home/fair/WWW/")

(defun top-directory (&optional filename)
  (setq filename (or filename (buffer-file-name)))
  (let ((prefixlen (length top-directory-prefix)))
    (if (or (> prefixlen (length filename))
	    (not (string= top-directory-prefix (substring filename 0 prefixlen))))
	(error "Can't recognize directory."))
    (let ((last prefixlen)
	  (result ""))
      (while (string-match "/" filename last)
	(setq last (match-end 0)
	      result (concat "../" result)))
      result)))

(defun process-indexees-for-file (&optional from-point)
  (interactive "P")
  (let (potential-begin-and-end
	potential-begin
	potential-end
	potential-string
	command-char
	adjust-beginning
	(temporary-ignore-list '())
	(dynamic-top-directory (top-directory))
	done)
    ;; This is FAIR-specific; avoid checking header and footer.
    ;; If FROM-POINT is non-nil, we assume that the menu bar is already hidden.
    (if (not from-point)
	(progn
	  (goto-char (point-min))
	  (if (search-forward "\n<hr>\n[ <a href=\"" nil t)
	      (narrow-to-region (point-min) (match-beginning 0)))
	  (goto-char (point-min))
	  (search-forward "</h1>\n\n")
	  (if (looking-at "By ") (forward-char 3))
	  (if (looking-at "EXTRA! \(") (search-forward "\n\n"))))
    (while (setq potential-begin-and-end (find-potential-indexee))
      (setq potential-begin (car potential-begin-and-end)
	    potential-end (cdr potential-begin-and-end))
      (setq command-char nil
	    adjust-beginning nil
	    done nil)
      (while (not done)
	(goto-char (if adjust-beginning
		       potential-begin
		     potential-end))
	(setq potential-string (clean-up-whitespace
				(buffer-substring potential-begin
						  potential-end)))
	(cond
	 ((member potential-string temporary-ignore-list)
	  (goto-char potential-end)
	  (setq done t))
	 ((and (not (member command-char '(?\C-a ?\C-e)))
	       (preclassify potential-begin potential-end potential-string))
	  (setq done t))
	 (t
	  (message "Classification for %s [hopqnx?]?" potential-string)
	  (setq command-char (read-char))
	  (cond ((= command-char ?\C-a)
		 (setq adjust-beginning t))
		((= command-char ?\C-e)
		 (setq adjust-beginning nil))
		((= command-char ?\C-f)
		 (if adjust-beginning
		     (if (= (1+ potential-begin) potential-end)
			 (beep)
		       (progn
			 (setq potential-begin (1+ potential-begin))
			 (if (and (= (char-after (1- potential-begin)) ?\")
				  (= (char-after potential-end) ?\"))
			     (setq potential-end (1+ potential-end)))))
		   (progn
		     (setq potential-end (1+ potential-end))
		     (if (and (= (char-after (1- potential-end)) ?\")
			      (= (char-after (1- potential-begin)) ?\"))
			 (setq potential-begin (1- potential-begin))))))
		((= command-char ?\C-b)
		 (if adjust-beginning
		     (progn
		       (setq potential-begin (1- potential-begin))
		       (if (and (= (char-after potential-begin) ?\")
				(= (char-after potential-end) ?\"))
			   (setq potential-end (1+ potential-end))))
		   (if (= (1+ potential-begin) potential-end)
		       (beep)
		     (progn
		       (setq potential-end (1- potential-end))
		       (if (and (= (char-after potential-end) ?\")
				(= (char-after potential-begin) ?\"))
			   (setq potential-begin (1+ potential-begin)))))))
		((= command-char ?\M-f)
		 (if adjust-beginning
		     (progn
		       (forward-word 2)
		       (backward-word 1)
		       (if (< (point) potential-end)
			   (setq potential-begin (point))
			 (beep)))
		   (progn
		     (if (looking-at "'s")
			 (forward-char 2))
		     (forward-word 1)
		     (if (looking-back-at "'s")
			 (backward-char 2))
		     (setq potential-end (point)))))
		((or (= command-char ?\M-b)
		     (= command-char 255)) ; M-DEL
		 (if adjust-beginning
		     (progn
		       (backward-word 1)
		       (setq potential-begin (point)))
		   (progn
		     (backward-word 2)
		     (forward-word 1)
		     (if (> (point) potential-begin)
			 (progn
			   (if (looking-back-at "'s")
			       (backward-char 2))
			   (setq potential-end (point)))
		       (beep)))))
		((= command-char ?x)
		 (goto-char potential-end)
		 (setq temporary-ignore-list (cons potential-string
						   temporary-ignore-list))
		 (setq done t))
		((member command-char '(?H ?O ?P ?Q))
		 (process-alias
		  potential-begin potential-end potential-string command-char)
		 (setq done t))
		((= command-char ?h)
		 (process-uncanonicalized-human
		  potential-begin potential-end potential-string)
		 (setq done t))
		((= command-char ?o)
		 (process-organization
		  potential-begin potential-end potential-string)
		 (setq done t))
		((= command-char ?p)
		 (process-uncanonicalized-publication
		  potential-begin potential-end potential-string)
		 (setq done t))
		((= command-char ?q)
		 (process-uncanonicalized-article
		  potential-begin potential-end potential-string)
		 (setq done t))
		((or (= command-char ?n)
		     (= command-char 127)) ; DEL
		 (goto-char potential-begin)
		 (forward-word 1)
		 (setq done t))
		((= command-char ?N)
		 (goto-char potential-end)
		 (setq done t))
		((= command-char ?s)
		 (end-of-line 1)
		 (setq done t))
		((= command-char ?!)
		 (if (y-or-n-p "Skip rest of file? ")
		     (progn
		       (goto-char (point-max))
		       (setq done t))))
		((= command-char ??)
		 (momentary-string-display indexing-keystroke-help
					   (window-start))
		 (read-char))
		(t
		 (beep))))))))
  (widen)
  (if (interactive-p)
      (message "Consider M-x write-index-obarray")))


(defun preclassify (begin end string)
  (let* ((sym (intern-soft string index-obarray))
	 (defs (symbol-value sym))
	 (classified nil))
    (while (and defs (not classified))
      (setq classified t)
      (let ((this-def (car defs)))
	(cond ((consp this-def)
	       (cond ((and (= ?H (car this-def))
			   (wi-y-or-n-p (format "Index %s as human %s? "
						string (cdr this-def))
					'(?h ?H)))
		      (process-canonicalized-human begin end (cdr this-def)))
		     ((and (= ?O (car this-def))
			   (wi-y-or-n-p (format "Index %s as organization %s? "
						string (cdr this-def))
					'(?o ?O)))
		      (process-organization begin end (cdr this-def)))
		     ((and (= ?P (car this-def))
			   (wi-y-or-n-p (format "Index %s as publication %s? "
						string (cdr this-def))
					'(?p ?P)))
		      (process-publication begin end (cdr this-def)))
		     ((and (= ?Q (car this-def))
			   (wi-y-or-n-p (format "Index %s as article titled %s? "
						string (cdr this-def))
					'(?q ?Q)))
		      (process-canonicalized-article begin end (cdr this-def)))
		     (t
		      (setq classified nil))))
	      ((and (= ?h this-def)
		    (wi-y-or-n-p (format "Index %s as human? " string)
				 '(?h ?H)))
	       (process-canonicalized-human begin end string))
	      ((and (= ?o this-def)
		    (wi-y-or-n-p (format "Index %s as organization? " string)
				 '(?o ?O)))
	       (process-organization begin end string))
	      ((and (= ?p this-def)
		    (wi-y-or-n-p (format "Index %s as publication? " string)
				 '(?p ?P)))
	       (process-publication begin end string))
	      ((and (= ?q this-def)
		    (wi-y-or-n-p (format "Index %s as article title? " string)
				 '(?q ?Q)))
	       (process-canonicalized-article begin end string))
	      (t
	       (setq classified nil))))
      (setq defs (cdr defs)))
    classified))

(defun process-alias (begin end string type)
  (let ((canonical
	 (let ((completion-ignore-case t))
	   (completing-read (format "%s is alias for %s: " string
				    (cond ((= ?H type) "human")
					  ((= ?O type) "organization")
					  ((= ?P type) "publication")
					  ((= ?Q type) "article title")))
			    index-obarray
			    (function
			     (lambda (elt)
			       (member (downcase type)
				       (symbol-value elt))))
			    nil
			    string))))
    (if (= ?\  (elt canonical (1- (length canonical))))
	(setq canonical (substring canonical 0 (1- (length canonical)))))
    (add-alias-entry string canonical type)
    (cond ((= ?H type) (process-canonicalized-human begin end canonical))
	  ((= ?O type) (process-organization begin end canonical))
	  ;; Should this be process-uncanonicalized-publication?
	  ((= ?P type) (process-publication begin end canonical))
	  ((= ?Q type) (process-canonicalized-article begin end canonical))
	  (t (error "What type is %c (%d)?" type type)))))

(defun add-alias-entry (alias canonical type)
  (add-nonalias-entry canonical (downcase type))
  (if (not (string= alias canonical))
      (let* ((sym (intern alias index-obarray))
	     (sym-val (and (boundp sym) (symbol-value sym)))
	     (new-entry (cons type canonical)))
	(if (not (member new-entry sym-val))
	    (set sym (cons new-entry sym-val))))))

(defun add-nonalias-entry (name type)
  (let* ((sym (intern name index-obarray))
	 (sym-val (and (boundp sym) (symbol-value sym))))
    (if (not (member type sym-val))
	(set sym (cons type sym-val)))))

(defun canonicalize-human-name (name)
  (if (string-match " [^ ]*$" name)
      (concat (substring name (1+ (match-beginning 0))) ", "
	      (substring name 0 (match-beginning 0)))))

(defun process-uncanonicalized-human (begin end name)
  (let* ((guess (canonicalize-human-name name))
	 (canonical-name
	  (if (and guess (wi-y-or-n-p (format "Index %s as %s? " name guess)))
	      guess
	    (read-string "Index format: " name))))
    (add-alias-entry name canonical-name ?H)
    (process-canonicalized-human begin end canonical-name)))

(defun process-canonicalized-human (begin end name)
  (add-nonalias-entry name ?h)
  (save-restriction
    (narrow-to-region begin end)
    (goto-char begin)
    (insert-index-ref-start-tag name)
    (goto-char (point-max))
    (insert "</a>")))

;; Or other proper name
(defun process-organization (begin end name)
  (add-nonalias-entry name ?o)
  (save-restriction
    (narrow-to-region begin end)
    (goto-char begin)
    (insert-index-ref-start-tag name)
    (goto-char (point-max))
    (insert "</a>")))

(defvar quote-punctuation-char-list '(?, ?. ?\; ?: ?! ??))
(defvar quote-punctuation-char-string ",.;:!?")

(defun process-uncanonicalized-publication (begin end name)
  ;; Is it more efficient to do the comparison on the string or the buffer?
  (if (and (= ?\" (char-after begin))
	   (= ?\" (char-after (1- end)))
	   (not (y-or-n-p (format "Do quotation marks belong around %s? " name))))
      (progn
	(save-excursion
	  (goto-char end)
	  (backward-delete-char 1)
	  (goto-char begin)
	  (delete-char 1))
	(setq end (- end 2)
	      name (substring name 1 (1- (length name))))
	(if (and (member (char-after (1- end)) quote-punctuation-char-list)
		 (not (y-or-n-p
		       (format "Does punctuation belong at end of %s? " name))))
	    (setq end (1- end)
		  name (substring name 0 (1- (length name)))))))
  (process-publication begin end name))



(defun process-publication (begin end name)
  (add-nonalias-entry name ?p)
  (save-restriction
    (narrow-to-region begin end)
    (goto-char begin)
    (insert-index-ref-start-tag name 'italic)
    (goto-char (point-max))
    (insert "</em></a>")))

(defun canonicalize-article-title (name)
  (if (and (string-match (concat "[" quote-punctuation-char-string "]\"$")
			 name)
	   (not (y-or-n-p (format "Does punctuation belong at end of %s? "
				  (if (> (length name) 28)
				      (concat "\"..." (substring name -24))
				    name)))))
      (setq name (concat (substring name 0 -2) "\"")))
  (surround-with-quotes name))

(defun surround-with-quotes (string)
  (if (not (= (aref string 0) ?\"))
      (setq string (concat "\"" string)))
  (if (not (= (aref string (1- (length string))) ?\"))
      (setq string (concat string "\"")))
  string)

(defun process-uncanonicalized-article (begin end name)
  (let* ((guess (canonicalize-article-title name))
	 (canonical-name
	  (surround-with-quotes
	   (if (and guess (wi-y-or-n-p (format "Index title as %s? " guess)))
	       guess
	     (read-string "Index format: " name)))))
    (add-alias-entry name canonical-name ?Q)
    (process-canonicalized-article begin end canonical-name)))


;; NAME is canonicalized (quote-surrounded, no extraneous punctuation),
;; but BEGIN and END aren't.
(defun process-canonicalized-article (begin end name)
  (add-nonalias-entry name ?q)
  ;; Purely a check.
  (if (not (string-match "^\".*\"$" name))
      (error "canonicalized article title isn't quoted: %s" name))

  (let ((end-mark (make-marker)))
    (set-marker end-mark end)
    (goto-char begin)
    (cond ((= (char-after (point)) ?\")
	   nil)
	  ((= (preceding-char) ?\")
	   (backward-char 1))
	  (t
	   (insert ?\")
	   (backward-char 1)))
    (insert-index-ref-start-tag (substring name 1 -1) 'quoted)
    (goto-char (marker-position end-mark))
    (set-marker end-mark nil)
    (cond ((= (preceding-char) ?\")
	   (let ((pre-quote-char (char-after (- (point) 2))))
	     ;; member test not strictly necessary.
	     ;; perhaps also check that everything up to here matches,
	     ;; or, better, remember answer to the question asked earlier.
	     (if (and (member pre-quote-char quote-punctuation-char-list)
		      (not (= pre-quote-char (aref name (- (length name) 2)))))
		 (progn
		   (backward-char 1)
		   (transpose-chars 1)
		   (backward-char 1)))))
	  ((= (char-after (point)) ?\")
	   (forward-char 1))
	  ((and (member (char-after (point)) quote-punctuation-char-list)
		(= (char-after (1+ (point))) ?\"))
	   ;; The user would have included the punctuation if it belonged.
	   (forward-char 1)
	   (transpose-chars 1)
	   (backward-char 1))
	  (t
	   (insert ?\")))
    (insert "</a>")))


(defun process-indexees-for-directory (directory &optional start-file)
  (interactive "DProcess indexees for directory: ")
  (let ((files (directory-files directory nil "\\.html$")))
    (while files
      (if (not (and start-file (string-lessp (car files) start-file)))
	  (progn
	    (find-file (expand-file-name (car files) directory))
	    (process-indexees-for-file)))
      (setq files (cdr files))))
  (message "Consider M-x write-index-obarray"))


;; This mirrors Perl routine strip_entry.
(defun index-file-letter (entry)
  (if (= ?\" (elt entry 0))
      (setq entry (substring entry 1)))
  (if (string-match "^\\(the\\|an?\\|la\\) \"?" entry)
      (setq entry (substring entry (match-end 0))))
  (if (string-match "^[`'$]" entry)
      (setq entry (substring entry (match-end 0))))
  (let ((letter (downcase (elt entry 0))))
    (cond ((and (>= letter ?a) (<= letter ?z))
	   letter)
	  ((string-match "^[1-9]" entry)
	   (cond ((string-match "^10[^0-9]" entry) ?t)
		 ((string-match "^11[^0-9]" entry) ?e)
		 ((string-match "^1[2-9]" entry)
		  (digit->first-letter (elt entry 1)))
		 (t (digit->first-letter (elt entry 0)))))
	  (t (error "What letter does this entry belong to? ")))))
;; (eq ?j (index-file-letter "La Journada"))
;; (eq ?t (index-file-letter "\"The Times\""))

(defun digit->first-letter (num-char)
  (let ((num-assoc (assoc num-char
			  '((?1 . ?o) (?2 . ?t) (?3 . ?t) (?4 . ?f) (?5 . ?f)
			    (?6 . ?s) (?7 . ?s) (?8 . ?e) (?9 . ?n)))))
    (if num-assoc
	(cdr num-assoc)
      (error "`%d' is not a digit." num-char))))

(defun insert-index-ref-start-tag (name &optional flag)
  (while (string-match "\"" name)
    (setq name (concat (substring name 0 (match-beginning 0))
		       (substring name (match-end 0)))))
  (insert "<a href=\"" dynamic-top-directory
	  ;; was "index-" ...
	  "index/" (index-file-letter name) ".html#" name
	  (cond ((eq flag 'italic)
		 "|i\"><em>")
		((eq flag 'quoted)
		 "|q\">")
		(t "\">"))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Reading and writing the index information
;;;

(defvar index-obarray-file "~/WWW/fair/index/raw-index.el"
  "File for the www-index obarray.")

;; Very slow, because pp-eval-expression is called on a huge list.
(defun write-index-obarray-slow ()
  (interactive)
  (let ((index-obarray-list '()))
    (mapatoms (function
	       (lambda (sym)
		 (setq index-obarray-list
		       (cons (cons (symbol-name sym) (symbol-value sym))
			     index-obarray-list))))
	      index-obarray)
    ;; (with-output-to-temp-buffer "obarray-output"
    ;;   (print index-obarray-list))
    (pp-eval-expression (list 'quote index-obarray-list)))
  (set-buffer "*Pp Eval Output*")
  (goto-char (point-min))
  (insert ";; Information about index entries for FAIR/EXTRA!\n")
  (goto-char (point-min))
  (write-file index-obarray-file)
  (kill-buffer (current-buffer))
  ;; For debugging, comment the kill-buffer and uncomment this.
  ;; (pop-to-buffer (current-buffer))
  )

;; This doesn't put the value on a separate line from the name, as the slow
;; version does.
(defun write-index-obarray-fast ()
  (interactive)
  (with-output-to-temp-buffer "*write-index-obarray-fast*"
    (princ ";; Information about index entries for FAIR/EXTRA!\n\(")
    (mapatoms (function
	       (lambda (sym)
		 (terpri) (princ " ")
		 (prin1 (cons (symbol-name sym) (symbol-value sym)))))
	      index-obarray)
    (princ "\n\)")
    (set-buffer "*write-index-obarray-fast*")
    (write-file index-obarray-file)
    ;; Why is this erring?
    ;; (kill-buffer (current-buffer))
    ;; For debugging, comment the kill-buffer and uncomment this.
    ;; (pop-to-buffer (current-buffer))
    ))

(fset 'write-index-obarray 'write-index-obarray-fast)

;; This should check that none of the false capitalizations appear.
(defun read-index-obarray ()
  (interactive)
  (if (file-exists-p index-obarray-file)
      (progn
	(setq index-obarray (make-vector 1023 0))
	(set-buffer (find-file-noselect index-obarray-file))
	(goto-char (point-min))
	(let ((index-obarray-list (read (current-buffer))))
	  (while index-obarray-list
	    (let ((this-entry (car index-obarray-list)))
	      (set (intern (car this-entry) index-obarray) (cdr this-entry))
	      (setq index-obarray-list (cdr index-obarray-list)))))
	(kill-buffer (current-buffer)))))

(if (not index-obarray)
    (read-index-obarray))

;; Testing
;; (set (intern "foo" index-obarray) '(1 2 3))
;; (set (intern "bar" index-obarray) '(4 5 6))
;; (set (intern "baz" index-obarray) '(7 8 9))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Utilities
;;;

;; Like y-or-n-p, but leaves cursor in buffer, not in minibuffer.
;; YES-RESPONSES is an additional list of characters
(defun wi-y-or-n-p (prompt &optional yes-responses)
  (if (or (member ?n yes-responses)
	  (member ?n yes-responses)
	  (member 127 yes-responses))
      (error "YES-RESPONSES must not contain n, N, or DEL."))
  (message (format "%s(y or n)" prompt))
  (let ((done nil)
	char)
    (while (not done)
      (setq char (read-char))
      (setq done (or (member char '(?Y ?y ?\  ?N ?n 127))
		     (member char yes-responses)))
      (if (not done)
	  (beep)))
    (or (member char '(?Y ?y ?\ ))
	(member char yes-responses))))
