#! /usr/local/bin/perl -ws

### Build HTML index
### Michael Ernst <mernst@theory.lcs.mit.edu>
### Last modified March 18, 1995

## This program does two things:
## 1. Create an index to a set of HTML documents.  The index consists
##    of 26 files (one per letter of the alphabet), each of which lists
##    terms, documents which index the term, and external links relevant
##    to the term.
##    * Index entries in HTML documents determine which terms are listed.
##    * The appropriate outside links are read from a file
##      (see read_external_links for details).
## 2. Comment or uncomment index entries.  Ordinarily each index entry
##    points to the appropriate portion of the index; but when the index for
##    that term contains only a single docuemnt, there is no point in making
##    the index entry point to the index, since the index contains no
##    additional information.  In that case, the index entry is commented
##    out, so that every active link in the document leads to new
##    information.  The index still contains the term and a pointer to the
##    document, even for commented-out index entries.
## Use the user-visible variables (below) to customize the program.


## To do:
##  * add a "concept" flag "|c" to index entries, make a smaller concept
##    index.  Add pointers to concept words from articles.
##  * entry_texts[$i] is always 1.  Why?  Should it be possible to be more?
##    If not, ditch some of the generality.
##  * Have separate tags for magazines and other publications, and in the
##    index, say "_Time_ magazine", not just "Time"
##  * Have "see" links for acronyms, etc.  also "seealso".
##    Probably do this by reading index entries and comparing with index name.
##  * write a script to remove (or comment out) non-concept index references
##    from articles and put them in a fair-nonindexed directory; add to tool
##    bar a method for going between indexed and nonindexed directories.
##  * read a configuration file to override user-visible variables.
##  * have a special tag when this is the author of the article.
##  * Make index entries indicate whether it's a concept entry (boldface?)
##    and, if not, the number of times it occurs.  (???)

###
### User-visible variables
###

$ROOT_DIRECTORY = "/u/mernst/WWW/fair/";
$INDEX_DIRECTORY = "/u/mernst/WWW/fair/index/";
# I could compute these, but this is more direct and easier for me just now.
$ROOT_RELATIVE_TO_INDEX = "../";
$INDEX_RELATIVE_TO_ROOT = "index/";

# This works because index.html is at top level
$ROOT_URL = "";
# $ROOT_URL = "http://theory.lcs.mit.edu/~mernst/fair/";

$EXTERNAL_LINKS = $INDEX_DIRECTORY . "outside-links";

## Associative array from issue to its URL
%issue_url = ('9412', '(<a href="9412/">12/94</a>)',
	      '9411', '(<a href="9411/">11/94</a>)',
	      'best-of-extra', '(<a href="best-of-extra/">Best of <em>EXTRA!</em></a>)',
	      'counterspin', '(<a href="counterspin/">CounterSpin</a>)',
	      'media-beat', '(<a href="media-beat/">Media Beat</a>)',
	      '', '');

###
### Internal variables
###

## Data structure:
##  A stripped entry is a unique top-level index entry, downcased,
##   with leading articles and (non-formatting) trailing flags removed,
##   with | changed to something earlier in the alphabet.
##  For each semistripped entry, we remember:
##   * full entry text, less modifiers
##   * formatting modifiers
##   * URL, if any
##   * list of:
##      * title
##      * URL
##      * issue (usually a date)

## Maybe add issue, for when article titles appear in index.

$entries = 0;

# arrays indexed by entry (an integer)
# These are 1-based arrays
@stripped_entry = ();		# downcased, less flags, leading articles, etc.
@entry_trailer = ();		# text inserted literally after entry
@entry_items = ();		# integer; number of items in 2D arrays
@entry_texts = ();		# integer
@entry_externals = ();		# integer

# Alternately, I could use the stripped entry as the index to all the arrays.
# This is probably more efficient, however; it reversed @stripped_entry.
%stripped_entry_index = ();	# from stripped entry to index

# two-dimensional arrays indexed by entry integer and by item integer
%entry_item_title = ();		# article title
%entry_item_url = ();		# article URL
%entry_item_issue = ();		# article issue in YYMM format
%entry_external = ();		# Whole text for external entry

# two-dimensional array indexed by entry integer and by entry_text
%entry_text = ();		# as it appears in the text, with modifiers

# Basically just a list of files in which indexing commands were found.
%indexed_files = ();
%non_indexed_files = ();

###
### The program.
###

$/ = "";		# Enable paragraph mode.
$* = 1;			# Enable multi-line patterns.
$[ = 1;			# Make arrays 1-based, so index = 0 means not found.

## To build the index:
##  * for each HTML file in the directory and its subdirectories:
##    * construct an appropriate index format entry
##    * scan file forward.  If we find
##      * index entry:  add it to a global list of such
##      * subheading that looks like a sidebar or sound-bites article:
##        modify the index format entry
##  * sort the list
##  * write the index file

&process_directory($ROOT_DIRECTORY);
&read_external_links($EXTERNAL_LINKS);
if ($hardcopy)
  { &write_hardcopy_index($INDEX_DIRECTORY);
    exit(0); }
&write_index_files($INDEX_DIRECTORY);
&process_indexed_files();
exit(0);

###########################################################################
### Fill the arrays
###

## To do:
##  issue date.
##  <index>

sub process_html_file {
  local($file) = @_;
  local($found_index_entry,$url,$url_sans_name,$title,$soundbites,$issue);
  local($entry,$entryletter,$stripped,$file_sans_root);
  $found_index_entry = 0;
  # print "process_html_file called on $file\n";

  # These can have index entries -- for the cover art, for instance.
  # # page without any index entries
  # if (m/(default|homepage|index|toc)\.html?/o)
  #   { return; }

  open(FILE, $file);
  $url = $file;
  $url =~ s/$ROOT_DIRECTORY/$ROOT_URL/o;
  $url_sans_name = $url;
  $_ = <FILE>;			# Read first line
  m/^<title>(.*)<\/title>/ || die "First line ($_) wasn't title in file $file";
  $title = $1;
  $soundbites = ($title eq "Soundbites");
  $issue = $url;
  $issue =~ s#(^|/)[^/]*$##;

#   ## Need to deal with Soundbites here.
#   # Deal with article title.
#   $stripped = &strip_entry($title);
#   if ($stripped eq "soundbites")
#     { $index = &add_entry_maybe($stripped, $title);
#       $item = ++$entry_items[$index];
#       $entry_item_title{$index, $item} = $title;
#       $entry_item_url{$index, $item} = $url;
#       $entry_item_issue{$index, $item} = $issue; }
#   elsif ($stripped_entry_index{$stripped})
#     { die "Found $stripped in stripped_entry_index" }
#   else
#     { $index = &add_entry_maybe($stripped, $title);
#       $entry_trailer[$index] = " " . &issue_to_html($issue); }

  while (<FILE>) {
    ## Check for new anchor_name and title
    if ($soundbites && m/<a name="(.*)">\n<h2>(.*)<\/h2>/)
      { $url = $url_sans_name . "#" . $1;
	$title = $2; };

    ## Look for index entries
    # 1, 5 = optional [ignored] "!-- ", " --"
    # 2 = path to root [ignored]
    # 3 = index letter
    # 4 = index entry
    while (m/<(!-- )?a\s+href=\"(\.\.\/)*index([-\/][a-z])?\.html?\#([^\">]*)\"( --)?>/go)
      { $entry = $4;
	$entryletter = $3;
	$stripped = &strip_entry($entry);
	# print "Found $entry";
	if (! defined($entryletter))
	  { ($file_sans_root = $file) =~ s/$ROOT_DIRECTORY//;
	    print "Reference to index.html should be index/",
		substr($stripped, $[, 1), ".html at $file_sans_root line $..\n"; }
	elsif (substr($entryletter, -1) cmp substr($stripped, $[, 1))
	  { ($file_sans_root = $file) =~ s/$ROOT_DIRECTORY//;
	    print "At $file_sans_root line $., index/", substr($stripped, $[, 1), ".html (not index$entryletter.html) should contain entry $entry\n"; }
	elsif (! ($entryletter =~ (m/^[-\/][a-z]$/)))
	  { print "Bad index letter $entryletter.\n"; }
	$index = &add_entry_maybe($stripped, $entry);
#	print "get_index_item $stripped=$index,$title,$url,$issue = ",&get_index_item($index, $title, $url, $issue),"\n";
	if (! &get_index_item($index, $title, $url, $issue))
  	  { $item = ++$entry_items[$index];
	    $entry_item_title{$index, $item} = $title;
	    $entry_item_url{$index, $item} = $url;
	    $entry_item_issue{$index, $item} = $issue; }
	if (! $found_index_entry)
	  { $found_index_entry = 1;
	    $indexed_files{$file} = 1; } }

    # If anchor conatains double-quote, complain
    if (m/<(!-- )?a\s+href=\"(\.\.\/)*index([-\/][a-z])?\.html?\#([^\">]*)\"([^>]*)\"( --)?>/go)
      { print "Oughtn't have quotes in index$3.html#$4\"$5"; }

    ## Look for uses of <index>
    ##

  }
  close(FILE);

  if (! $indexed_files{$file})
    { $non_indexed_files{$file} = 1; }
}

sub process_directory {
  local($directory) = @_;
  
  print "Processing directory $directory...\n";

  opendir(DIR, $directory) || die "Can't open $directory";
  local(@filenames) = readdir(DIR);
  closedir(DIR);
  
  for (@filenames)
  { # print "Processing $_...\n";
    next if $_ eq '.';
    next if $_ eq '..';
    $_ = $directory . $_;
    if (-l $_)
      { }			# do nothing if link to another dir.
    elsif (-d $_)
      { &process_directory($_ . '/'); }
    elsif (m/\.html?$/o)
      { &process_html_file($_); } }
}

sub get_index_item {
  local($index, $title, $url, $issue) = @_;
  local($i);
  for ($i = 1; $i <= $entry_items[$index]; $i++)
    { if (($title eq $entry_item_title{$index, $i})
	  && ($url eq $entry_item_url{$index, $i})
	  && ($issue eq $entry_item_issue{$index, $i}))
        { return $i; } }
  # default return value
  0;
}

sub add_entry_maybe {
  local($stripped, $title) = @_;
  local($index);

  if ($index = $stripped_entry_index{$stripped})
    { return $index; }

  $entries++;
  $stripped_entry_index{$stripped} = $entries;
  $stripped_entry[$entries] = $stripped;
  &add_entry_item($entries, $title);
  $entry_items[$entries] = 0;
  #   # Don't do this.
  #   $entry_texts[$entries] = 0;
  #   $entry_externals[$entries] = 0;
  $entries;
}

sub add_entry_item {
  local($index, $title) = @_;
  local($i);
  for ($i = 1; $i <= $entry_texts[$index]; $i++)
    { if ($entry_text{$index, $i} eq $title)
        { return; } }
  $entry_text{$index,++$entry_texts[$index]} = $title;
}

sub strip_entry
{ local($entry) = @_;
  $entry =~ tr/A-Z/a-z/;
  $entry =~ s/^"//;
  $entry =~ s/^(the|an?) "?//;
  $entry =~ s/^la //;		# Can I always strip leading "La "?
  $entry =~ s/^\$//;
  $entry =~ s/'//g;		# for names like O'Beirne, D'Amato, etc.
  $entry =~ s/`//g;		# for quotations (the above helps too)
  $entry =~ s/^de /de/;		# for proper names: de Borchgrave
  $entry =~ s/^du /du/;		# for proper names: Du Pont
	# Do something about "L.A."?
  $entry =~ s/\|(.*)[c](.*)/|$1$2/;
  $entry =~ s/\|(.*)[o](.*)/|$1$2/;
  $entry =~ s/\|/!/;		# ! appears before letters in ASCII
	# Don't get rid of all spaces:  index "foo bar" before "fooable"

  # Special cases for items I know occur in the index; add more as they appear
  $entry =~ s/^11([^0-9])/eleven$1/;
  $entry =~ s/^2([^0-9])/two$1/;
  $entry =~ s/^20([^0-9])/twenty$1/;
  $entry =~ s/^21st /twenty-first/;
  $entry =~ s/^35([^0-9])/thirty-five$1/;
  $entry =~ s/^4([^0-9])/four$1/;
  $entry =~ s/^60([^0-9])/sixty$1/;
  $entry =~ s/^800([^0-9])/eight hundred$1/;
  $entry =~ s/^90'?s/nineties/;	# apostrophe already stripped, above
  $entry =~ s/ \$treet/ street/;
  $entry;
}


###########################################################################
### Read external links
###

## The format of the file is alternating groups of entries and links.
## Each entry is preceded by no whitespace.
## Each link is preceded by whitespace.
## In a group, each link will appear in the index under each entry.
## Lines starting with # (possibly preceded by whitespace) are ignored

sub read_external_links
{ local ($file) = @_;
  local ($reading_entries,@entry_indices,$stripped,$index);

  $reading_entries = 1;
  @entry_indices = ();

  $/ = "\n";
  $* = 0;

  open(FILE, $file);

  while (<FILE>) {
    # print $reading_entries,' ',$_;
    chop;
    if (m/^\s*#/)
      { }
    elsif (m/^[ \t]/)
      { $reading_entries = 0;
	s/[ \t]*$//;		# strip trailing spaces
	if (@entry_indices)
	  { foreach $index (@entry_indices)
	      { $entry_external{$index,++$entry_externals[$index]} = $_; } }
	else
	  { print "No entry for external link $_\n"; } }
    elsif (m/^\s*$/)
      { $reading_entries = 0; }
    else # m/^[^ \t]/
      { if (! $reading_entries)
	  { $reading_entries = 1;
	    @entry_indices = (); }
	$stripped = &strip_entry($_);
	$index = $stripped_entry_index{$stripped};
	if ($index)
  	  { push(@entry_indices,$index); }
	else
	  { print "No index entry $_ (stripped: $stripped)\n"; } }
    }
  close(FILE);
  
  $/ = "";
  $* = 1;
}


###########################################################################
### Process the filled arrays
###

sub format_index_entry
{ local($term) = @_;
  local($entry, $modifiers) = split(/\|/, $term);
  if (! defined($modifiers))
    { return $entry; }
  if (index($modifiers,'q') != $[ - 1)
    { $entry = '"' . $entry . '"'; }
  if (index($modifiers,'i') != $[ - 1)
    { $entry = '<em>' . $entry . '</em>'; }
  $entry;
}

sub issue_to_html
{ local($issue) = @_;
  $issue_url{$issue}; }

sub make_index_entry_i_j
{ local($i, $j) = @_;
  &make_index_entry($entry_item_title{$i,$j},$entry_item_url{$i,$j},$entry_item_issue{$i,$j}); }

sub make_index_entry
{ local($title, $url, $issue) = @_;
  "<a href=\"${ROOT_RELATIVE_TO_INDEX}$url\">$title</a> " . &issue_to_html($issue); }

sub write_index_files
{ local($directory) = @_;
  local($i,$j,$indexletter,$total_items);

  if ($directory =~ m/\/$/)
    { chop($directory); }

  #   if (-e "$directory/index.html")
  #     { system "mv $directory/index.html $directory/index.html.bak"; }
  for ($i = "a"; $i le "z"; $i++)
    { if (-e "$directory/$i.html")
      { system "mv $directory/$i.html $directory/$i.html.bak"; } } 

  $total_items = 0;
  for ($entry = 1; $entry <= $entries; $entry++)
    { $total_items += $entry_items[$entry]; }
  print "Writing $entries entries, $total_items items.\n";

  @s = (1 .. $entries);
  ## print @s;
  @s = sort { $stripped_entry[$s[$a]] cmp $stripped_entry[$s[$b]] } @s;

  ## print @s;
  ## for ($i = 1; $i <= $entries; $i++)
  ##  { print $s[$i],$entry[$s[$i]],"\n"; }

  $indexletter = "a";
  &startindex($indexletter,$directory);

  foreach $i (@s)
    { if ($indexletter cmp substr($stripped_entry[$i], $[, 1))
      { &stopindex($indexletter);
	$indexletter = substr($stripped_entry[$i], $[, 1);
	&startindex($indexletter,$directory); }

      # print "entry_texts[$i] = $entry_texts[$i]\n";

      for ($j = 1; $j <= $entry_texts[$i]; $j++)
        { print INDEX "\n <a name=\"$entry_text{$i,$j}\">"; }

      ## I ought to check that all the entry_texts are compatible.
      print INDEX "\n <dt>";
      print INDEX &format_index_entry($entry_text{$i,$[});
      print INDEX $entry_trailer[$i];

      @ss = (1 .. $entry_items[$i]);
      @ss = sort { ($entry_item_issue{$i, $ss[$a]} 
  		      cmp $entry_item_issue{$i, $ss[$b]})
		   || ($entry_item_title{$i, $ss[$a]}
		         cmp $entry_item_title{$i, $ss[$b]}) } @ss;
      foreach $j (@ss)
        { # print INDEX "\n  <dd>",&make_index_entry($entry_item_title{$i,$j},$entry_item_url{$i,$j},$entry_item_issue{$i,$j});
	  print INDEX "\n  <dd>",&make_index_entry_i_j($i,$j); }
      if ($entry_externals[$i])
        { # print "entry_externals[$i] = $entry_externals[$i]\n";
	  print INDEX "\n  <dd>External links:\n   <ul>";
	  for ($j = 1; $j <= $entry_externals[$i]; $j++)
	    { print INDEX "\n    <li>",$entry_external{$i,$j}; }
	  print INDEX "\n   </ul>"; }
      }

  &stopindex($indexletter);
}


sub startindex
{ local($indexletter,$directory) = @_;

  if ($directory =~ m/\/$/)
    { chop($directory); }

  open (INDEX, ">$directory/$indexletter.html")
      || die "Can't open $directory/$indexletter.html for output.";

  print INDEX <<EndOfIndexHeaderPart1;
<title>Index, part \u$indexletter, for FAIR publications</title>
<link rev="made" href="mailto:mernst@theory.lcs.mit.edu">

<h1>Index, part \u$indexletter, for FAIR publications</h1>

EndOfIndexHeaderPart1

  &indexmenu($indexletter);

  print INDEX <<EndOfIndexHeaderPart2;
Help improve this index!  Please <a
href="http://www.ai.mit.edu/people/mernst/fair/send-comment-www.html">suggest</a>
external pages to point to, or additional index entries to add.<p>


EndOfIndexHeaderPart2
  print INDEX "<dl>";		# no linefeed

  # no return value
}

sub indexmenu
{ local($indexletter) = @_;
  local($i);

  # we can't use $i le "z" because in perl, "z"++ = "aa".
  for ($i = "a"; length($i)==1; $i++)
    { if ($i eq $indexletter)
        { print INDEX "\u$i \n"; }
      else
        { print INDEX "<a href=\"$i.html\">\u$i</a> \n"; } }
  print INDEX "<p>\n\n";
}

sub stopindex
{ local($indexletter) = @_;

  print INDEX "\n</dl>\n\n";
  &indexmenu($indexletter);

# "Up" could be index.html, but that doesn't have any content not on the
# index/<letter> pages, so don't bother.
  print INDEX <<EndOfIndexTrailer;
<hr>
[ <a href="$ROOT_RELATIVE_TO_INDEX">Up</a> | 
<a href="$ROOT_RELATIVE_TO_INDEX">Top</a> | 
<a href="http://www.ai.mit.edu/people/mernst/fair/send-comment-www.html">Mail/Suggest</a> ]<br>
EndOfIndexTrailer

  close(INDEX);
}  


sub write_hardcopy_index
{ local($directory) = @_;
  local($i,$j,$indexletter,$total_items);

  if ($directory =~ m/\/$/)
    { chop($directory); }

  if (-e "$directory/hardcopy_index")
    { system "mv $directory/hardcopy_index $directory/hardcopy_index.bak"; }

  $total_items = 0;
  for ($entry = 1; $entry <= $entries; $entry++)
    { $total_items += $entry_items[$entry]; }
  print "Writing $entries entries, $total_items items.\n";

  @s = (1 .. $entries);
  ## print @s;
  @s = sort { $stripped_entry[$s[$a]] cmp $stripped_entry[$s[$b]] } @s;

  ## print @s;
  ## for ($i = 1; $i <= $entries; $i++)
  ##  { print $s[$i],$entry[$s[$i]],"\n"; }

  open (INDEX, ">$directory/hardcopy_index")
    || die "Can't open $directory/hardcopy_index.html for output.";

  print INDEX "Index for FAIR publications\n\n\n";

  foreach $i (@s)
    { 
      ## I ought to check that all the entry_texts are compatible.
      print INDEX "\n",&un_html(&format_index_entry($entry_text{$i,$[}));

      @ss = (1 .. $entry_items[$i]);
      @ss = sort { ($entry_item_issue{$i, $ss[$a]} 
  		      cmp $entry_item_issue{$i, $ss[$b]})
		   || ($entry_item_title{$i, $ss[$a]}
		         cmp $entry_item_title{$i, $ss[$b]}) } @ss;
      foreach $j (@ss)
        { # print INDEX "\n  ",&un_html(&make_index_entry_i_j($i,$j));
	  print INDEX "\n  ",&un_html(&make_index_entry($entry_item_title{$i,$j},$entry_item_url{$i,$j},$entry_item_issue{$i,$j})); } }

  close(INDEX);
}

sub un_html
{ local($string) = @_;
  $string =~ s/<[^>]*>//g;
  $string;
}


###########################################################################
### Rewrite files with indexing commands
###

sub process_indexed_files
{
  local($file);

  foreach $file (keys(%indexed_files))
    { &process_indexed_file($file); }
  foreach $file (keys(%non_indexed_files))
    { if ((! ($file =~ m/index[-\/][a-z]\.html$/))
	  && (! ($file =~ m/\/(back-links|credits|letter-writing|limbaugh-response|whats-new).html$/)))
        { print "No indexing found in $file\n"; } }
}

# Problem: this doesn't respect spacing between paragraphs: 3 or more
# newlines are turned into 2 newlines.
sub process_indexed_file
{ local($file) = @_;
  local($file_url,$url_sans_name,$bak,$soundbites);

  $file_url = $file;
  $file_url =~ s/$ROOT_DIRECTORY/$ROOT_URL/o;
  $url_sans_name = $file_url;

  $bak = $file . '.bak';
  rename($file, $bak);
  open(FILE, $bak);
  open(OUTFILE, ">$file");

  $_ = <FILE>;			# Read first paragraph
  $soundbites = m/^<title>Soundbites<\/title>/;
  print OUTFILE;

  while (<FILE>) {

    ## Check for new anchor_name and title
    if ($soundbites && m/<a name="(.*)">\n<h2>(.*)<\/h2>/)
      { $file_url = $url_sans_name . "#" . $1; };

    # 1,5,8,9 = optional [ignored] "!-- ", " --", "!-- ", " --"
    # 2 = url (3 = path to root, 4 = index entry)
    # 6 = anchor text (also 7, probably, but I'm paranoid)
    s@<(!-- )?a\s+href=\"((\.\./)*index[-/][a-z]\.html?\#([^\">]*))\"( --)?>(([^/]|/[^a])*)<(!-- )?/a( --)?>@&comment_or_uncomment_index_reference($4,$2,$6,$file_url)@ge;

#   ## Look for index entries
#   while (m/<(!-- )?(a\s+href=\"(\.\.\/)*index\.html?\#([^\">]*)\")( --)?>/go)
#     { $entry = $4;
#       $stripped = &strip_entry($entry);
#       $index = $stripped_entry_index{$stripped}
# 		|| die("Couldn't find $stripped in stripped_entry-index");
#       if ($entry_items[$index] == 1)
# 	{ $1 = "";
# 	  $5 = ""; }
#       else
# 	{ $1 = "!-- ";
# 	  $5 = " --"; } }

    print OUTFILE; }
  close(FILE);
  close(OUTFILE);
  @stats=stat($bak);
  # Add 1 to let other programs (like Emacs, httpd) see that the file is modified
  # 8 + $[ = atime, 9 + $[ = mtime
  utime @stats[8 + $[] + 1, @stats[9 + $[] + 1, $file;

}

sub comment_or_uncomment_index_reference
{ local($entry,$url,$anchor,$file_url) = @_;
  local($stripped,$index);
  # print "com_or_uncom($entry,$url,$anchor)\n";
  $stripped = &strip_entry($entry);
  $index = $stripped_entry_index{$stripped}
            || die("Couldn't find $stripped in stripped_entry-index");
#   print $stripped, " ", $entry_items[$index], "\n";
#   for ($i = 1; $i <= $entry_items[$index]; $i++)
#     { print "   ", $entry_item_title{$index, $i}, "+", $entry_item_title{$index, $i}, "+", $entry_item_url{$index, $i}, "\n"; }
  if (($entry_externals[$index] == 0)
      && ($entry_items[$index] == 1)
      && ($entry_item_url{$index,1} eq $file_url))
    { "<!-- a href=\"$url\" -->$anchor<!-- /a -->" }
  else
    { "<a href=\"$url\">$anchor</a>" }
}


###########################################################################
### Rewrite indexing commands
###

## Rewrite all .html files.
# &rewrite_index_commands_directory($ROOT_DIRECTORY);
# &rewrite_index_commands_directory("/u/mernst/tmp/");

sub rewrite_index_commands_directory {
  local($directory) = @_;
  
  print "Processing directory $directory...\n";

  opendir(DIR, $directory) || die "Can't open $directory";
  local(@filenames) = readdir(DIR);
  closedir(DIR);
  
  $/ = "\n";
  $* = 0;

  for (@filenames)
  { # print "Considering $_...\n";
    next if $_ eq '.';
    next if $_ eq '..';
    $_ = $directory . $_;
    if (-l $_)
      { }			# do nothing if link to another dir.
    elsif (-d $_)
      { &rewrite_index_commands_directory($_ . '/'); }
    elsif (m/\.html?$/o)
      { &rewrite_index_commands_html_file($_); } }

  $/ = "";
  $* = 1;
}

sub rewrite_index_commands_html_file
{ local($file) = @_;
  # print "Processing file $file...\n";
  
  $bak = $file . '.bak';
  rename($file, $bak);
  open(FILE, $bak);
  open(OUTFILE, ">$file");

  while (<FILE>) {

    # ## Replace "index.html#foo" with "index/f.html#foo".
    # # 1 = path to root, 2 = [ignored] dash and index letter, 3 = index entry
    # s@href=\"(\.\./)*index([-/][a-z])?\.html?\#([^\">]*)\"@&rewrite_index_commands_reference($1,$3)@ge;

    ## Replace "index-f.html" with "index/f.html".
    s@index-([a-z])\.html@index/\1.html@g;

    print OUTFILE; }
  close(FILE);
  close(OUTFILE);
  @stats=stat($bak);
  # 8 + $[ = atime, 9 + $[ = mtime
  utime @stats[8 + $[], @stats[9 + $[], $file;

}

sub rewrite_index_commands_reference
{ local($path,$entry) = @_;
  local($stripped);
  if (! (defined($path))) { $path = ""; }
  # print "rewrite_index_commands_reference($path,$entry)\n";
  $stripped = &strip_entry($entry);
  # return value
  "href=\"${path}index/" . substr($stripped, $[, 1) . ".html#$entry\"";
}


###########################################################################
### Silence compiler warnings
###

if ($foo && !$foo)
  { &comment_or_uncomment_index_reference();
    &rewrite_index_commands_reference(); 
    $INDEX_RELATIVE_TO_ROOT;
    $hardcopy; }
